from ..service import ReceivingService, ThreadService
from ..wire import RawWire

class SaveWav(ThreadService, ReceivingService):
    input_wire_class=RawWire
    
    def __init__(self, filename=None):
        super(SaveWav, self).__init__()

        if filename is None:
            filename="data.wav"
        
        self.outputfile=filename
        
    def start_process(self):
        global wav
        
        try:
          import wav
        except Exception as ex:
            self.print_message( "import error: {0}".format(str(ex)))

        self.output=wave.open(self.outputfile, 'wb')
        self.output.setnchannels(1)
        if self.input_wire.FORMAT!="int16":
            raise Exception("Format to be saved in wav needs to be int16")
        self.output.setsampwidth(2)
        self.output.setframerate(self.input_wire.RATE)
        
        super(SaveWav, self).start_process()

    def process(self, data):
        self.output.writeframes(data)

    def cleanup(self):
        self.output.close()



from ..service import ReceivingService
from ..wire import RawWire

try:
    import pyaudio
    HAS_PYAUDIO=True

    pyaudio_format=dict(int16=pyaudio.paInt16, float32=pyaudio.paFloat32)
    pyaudio_nbytes=dict(int16=2, float32=4)
  
except ImportError:
    HAS_PYAUDIO=False


class PyAudioPlay(ReceivingService):
    input_wire_class=RawWire
    def __init__(self, frames_per_buffer=2048, output_device_index=None):
        if not HAS_PYAUDIO:
          raise Exception("pyaudio module not or not correctly installed")        
        self.pyaudio=pyaudio.PyAudio()
        self.player=None
        self.frames_per_buffer=frames_per_buffer
        self.output_device_index=output_device_index
        super(PyAudioPlay, self).__init__()

    def _callback(self, in_data, frame_count, time_info, status):
        data=self.receive_input(block=False)
        if data is None:
            self.print_message("playback buffer underrun")
            data=bytes(bytearray(pyaudio_nbytes[self.input_wire.FORMAT]*frame_count))
        return (data, pyaudio.paContinue)

    def start(self):
        self.player = self.pyaudio.open(
                format=pyaudio_format[self.input_wire.FORMAT],
                channels=self.input_wire.CHANNELS,
                rate=self.input_wire.RATE,
                frames_per_buffer=self.frames_per_buffer,
                output=True,
                stream_callback=self._callback,
                output_device_index=self.output_device_index
                )
                
    def stop(self):
        if self.player:
            self.player.stop_stream()

    def close(self):
        if self.player:
            self.player.stop_stream()
            self.player.close()

        self.pyaudio.terminate()
        



from ..service import ThreadService, SourceService, ReceivingService
from ..wire import PulseWire, HistogramWire

import numpy
import pickle

class AggregateHistogram(ThreadService,ReceivingService, SourceService):
    input_wire_class=PulseWire
    output_wire_class=HistogramWire
    def __init__(self, nchannels=100, vmin=0, vmax=1., outfile="histogram"):
        super(AggregateHistogram, self).__init__()

        self.vmin=vmin
        self.vmax=vmax
        self.nchannels=nchannels
        self.outfile=outfile

        self.hist, self.bins=numpy.histogram([], bins=self.nchannels, 
            range=(self.vmin,self.vmax))

    def output_protocol(self, wire):
        super(AggregateHistogram, self).output_protocol(wire)
        wire.nchannels=self.nchannels
        wire.vmin=self.vmin
        wire.vmax=self.vmax
        wire.unit=self.input_wire.unit

    def process(self, data):
        signal=[x[1] for x in data]
        hist, bins=numpy.histogram(signal, bins=self.nchannels, 
            range=(self.vmin,self.vmax))
        
        self.hist=self.hist+hist

        return dict(hist=self.hist, bins=self.bins)

    def cleanup(self):
        if self.outfile is not None:
            f=open(self.outfile+".pkl","wb")
            pickle.dump((self.hist, self.bins),f)
            f.close()
        super(AggregateHistogram, self).cleanup()



#!/usr/bin/env python
"""
gammalab-histogram makes a histogram of detected peak heights. 
"""

import argparse

try:
  import matplotlib
  matplotlib.use('TkAgg')
except:
  pass

from gammalab import main
from gammalab.acquisition import SoundCard, FileReplay
from gammalab.transform import Raw2Float, SecondOrder
from gammalab.analysis import PulseDetection, FittedPulseDetection
from gammalab.analysis import AggregateHistogram
from gammalab.analysis import Count
from gammalab.backend import PlotHistogram

def run(threshold=0.003, nchannels=500, vmax=2000., offset=0, scale=5400., 
        drift=0., runtime=None, outfile=None, log=True, do_plot=True,
        input_device_name="", inputfile=None, realtime=True,
        fitpulse=False, fit_threshold=0.95, raw_values=False):

    if raw_values:
        print("[Gammalab-histogram] check detector output! we assume range of [0,1]")
        scale=1
        vmax=1

    if inputfile is not None:
        source=FileReplay(filename=inputfile, realtime=realtime)
    else:
        source=SoundCard(input_device_name=input_device_name)
    convert=Raw2Float()
    if fitpulse:
        detect=FittedPulseDetection(threshold=threshold, fit_threshold=fit_threshold)
    else:
        detect=PulseDetection(threshold=threshold)
    
    count=Count(outfile=outfile+".counts" if outfile is not None else None)
    histogram=AggregateHistogram(nchannels=int(nchannels*scale/vmax),
                                 vmin=0, 
                                 vmax=vmax, 
                                 outfile=outfile+".histogram" if outfile is not None else None)
    
    source.plugs_into(convert)
    convert.plugs_into(detect)
    detect.plugs_into(count)

    if raw_values:
        detect.plugs_into(histogram)
    else:
        calibrate=SecondOrder(offset=offset, scale=scale, drift=drift)
        detect.plugs_into(calibrate)
        calibrate.plugs_into(histogram)

    if do_plot:
        plothistogram=PlotHistogram(xmin=0, 
                                    xmax=vmax, 
                                    outfile=outfile,
                                    log=log)
        histogram.plugs_into(plothistogram)

    main(runtime)

def new_argument_parser():
    "Parse command line arguments"
    parser = argparse.ArgumentParser( formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                      description=__doc__)
    parser.add_argument(
        '--threshold',
        dest='threshold',
        default=0.003,
        type=float,
        help='(noise) threshold for pulse detection',
    )
    parser.add_argument(
        '--nchannels',
        dest='nchannels',
        default=100,
        type=int,
        help='number of channels per MeV (bins) in the histogram',
    )
    parser.add_argument(
        '--vmax',
        dest='vmax',
        default=2000,
        type=float,
        help='maximum energy bin for histogram',
    )
    parser.add_argument(
        '--offset',
        dest='offset',
        default=0.,
        type=float,
        help='signal energy offset (in keV)',
    )
    parser.add_argument(
        '--scale',
        dest='scale',
        default=5400,
        type=float,
        help='signal energy scale (in keV)',
    )
    parser.add_argument(
        '--drift',
        dest='drift',
        default=0,
        type=float,
        help='second order correction parameter (fractional deviation at raw value 1)',
    )
    parser.add_argument(
        '--outfile',
        dest='outfile',
        default=None,
        help='output file root',
    )
    parser.add_argument(
        '--runtime',
        dest='runtime',
        default=None,
        type=float,
        help='runtime in seconds',
    )
    parser.add_argument(
        '--log',
        dest='log',
        action="store_true",
        help='plot logarithmic y-axis',
    )
    parser.add_argument(
        '--do-plot',
        dest='do_plot',
        action="store_true",
        help='show plot',
    )
    parser.add_argument(
        '--input_device_name',
        dest='input_device_name',
        default="",
        type=str,
        help='select input device by (fuzzy matched) name',
    )
    parser.add_argument(
        '--infile',
        dest='inputfile',
        default=None,
        help='(optional) input file',
    ) 
    parser.add_argument(
        '--realtime',
        dest='realtime',
        action="store_true",
        help='file input in realtime',
    )       
    parser.add_argument(
        '--fitpulse',
        dest='fitpulse',
        action="store_true",
        help='use pulse fitting for amplitude?',
    )       
    parser.add_argument(
        '--fit-threshold',
        dest='fit_threshold',
        default=0.95,
        type=float,
        help='minimum peak to determine amplitude through fitting',
    )
    parser.add_argument(
        '--raw',
        dest='raw_values',
        action="store_true",
        help='histogram of raw detector values (no calibration or scaling)',
    )    

    return parser.parse_args()

if __name__=="__main__":
    args = new_argument_parser()
    run(**vars(args))




from setuptools import setup, find_packages

setup(
    name='GammaLab',
    version='0.5.0',
    description='Toolkit for soundcard gamma ray spectroscopy',
    author='F.I. Pelupessy',
    url='https://github.com/ipelupessy/gammalab',
    license='GPLv3',
    packages=find_packages('src'),
    package_data={},
    package_dir={'' : 'src'},
    install_requires=['wheel', 'numpy', 'SoundCard'],
    python_requires='>=3.6',
    scripts=['bin/gammalab-histogram.py', 'bin/gammalab-monitor.py','bin/gammalab-calibration-calc.py'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3) ',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.10',
        'Topic :: Scientific/Engineering :: Physics',
    ],
    long_description=open('README.rst').read(),
)



from . import all_services, shared_output

import multiprocessing
import threading
import time

try:
    input=raw_input
except:
    pass

def output_thread():
    message=None
    while message!="[Shutdown] done":
        message=shared_output.get(True)
        print(message)

def startup():
    print("[Startup] entry")

    t=multiprocessing.Process(target=output_thread, name="output")
    t.daemon=True
    t.start()

    for s in all_services:
        s.start()

def shutdown():
    if all_services:
        shared_output.put("[Shutdown] shutting down")

    for s in all_services:
          s.stop()

    while all_services:
        s=all_services.pop()
        try:
          s.close()
        except Exception as ex:
          shared_output.put("[Shutdown] "+str(ex))

    shared_output.put("[Shutdown] done")

    exit(0)

def main(timeout=None):

    startup()

    if timeout is not None:
        timer=threading.Timer(timeout, shutdown)
        timer.daemon=True
        timer.start()
        
    shared_output.put("[Main] running")

    input()

    shutdown()
    
    
